/*************************************************
* BER Decoding Source File                       *
* (C) 1999-2002 The Botan Project                *
*************************************************/

#include <botan/ber_dec.h>

namespace Botan {

namespace {

/*************************************************
* Check an object's type and size                *
*************************************************/
void check_object(const BER_Object& object, ASN1_Tag tag,
                  const std::string& type, u32bit length = 0,
                  bool check_length = false)
   {
   if(object.tag != tag)
      throw Decoding_Error("BER: Tag mismatch: Expecting " + type + ", got " +
                           to_string(object.tag));
   if(check_length && object.value.size() != length)
      throw Decoding_Error("BER decoded a " + type + " of incorrect size");
   }

}

namespace BER {

/*************************************************
* Decode a BER encoded NULL                      *
*************************************************/
void decode_null(BER_Decoder& decoder)
   {
   BER_Object object = decoder.get_next_object();
   check_object(object, NULL_TAG, "NULL", 0, true);
   }

/*************************************************
* Decode a BER encoded BOOLEAN                   *
*************************************************/
void decode(BER_Decoder& decoder, bool& boolean)
   {
   BER_Object object = decoder.get_next_object();
   check_object(object, BOOLEAN, "BOOLEAN", 1, true);
   boolean = (object.value[0]);
   }

/*************************************************
* Decode a small BER encoded INTEGER             *
*************************************************/
void decode(BER_Decoder& decoder, u32bit& out)
   {
   BigInt integer;
   decode(decoder, integer);
   out = integer.to_u32bit();
   }

/*************************************************
* Decode a BER encoded INTEGER                   *
*************************************************/
void decode(BER_Decoder& decoder, BigInt& integer)
   {
   BER_Object object = decoder.get_next_object();
   check_object(object, INTEGER, "INTEGER");

   integer = 0;
   if(object.value.size() == 0)
      return;

   const bool negative = (object.value[0] & 0x80);

   if(negative)
      {
      for(u32bit j = object.value.size(); j > 0; j--)
         if(object.value[j-1]--)
            break;
      for(u32bit j = 0; j != object.value.size(); j++)
         object.value[j] = ~object.value[j];
      }

   integer = BigInt(object.value, object.value.size());

   if(negative)
      integer.flip_sign();
   }

/*************************************************
* BER decode a BIT STRING or OCTET STRING        *
*************************************************/
void decode(BER_Decoder& decoder, SecureVector<byte>& buffer, ASN1_Tag expect)
   {
   BER_Object object = decoder.get_next_object();
   if(expect == OCTET_STRING)
      check_object(object, OCTET_STRING, "OCTET STRING");
   else if(expect == BIT_STRING)
      check_object(object, BIT_STRING, "BIT STRING");
   else
      throw Invalid_Argument("BER decoding: must expect BIT or OCTET STRING");

   if(object.tag == OCTET_STRING)
      buffer = object.value;
   else
      {
      if(object.value[0])
         throw Decoding_Error("BER: BIT STRING is not a multiple of 8 bits");
      buffer.set(object.value + 1, object.value.size() - 1);
      }
   }

/*************************************************
* Decode a BER encoded OID                       *
*************************************************/
void decode(BER_Decoder& decoder, OID& oid)
   {
   BER_Object object = decoder.get_next_object();
   check_object(object, OBJECT_ID, "OBJECT IDENTIFIER");

   if(object.value.size() < 2)
      throw Decoding_Error("BER decoding error: OID encoding too short");

   oid += (object.value[0] / 40);
   oid += (object.value[0] % 40);

   u32bit j = 0;
   while(j != object.value.size() - 1)
      {
      u32bit component = 0;
      while(j != object.value.size() - 1)
         {
         j++;
         component = (component << 7) + (object.value[j] & 0x7F);
         if(!(object.value[j] & 0x80))
            break;
         }
      oid += component;
      }
   }

/*************************************************
* Decode and return a BER encoded SEQUENCE       *
*************************************************/
BER_Decoder get_subsequence(BER_Decoder& decoder)
   {
   BER_Object object = decoder.get_next_object();
   check_object(object, ASN1_Tag(SEQUENCE | CONSTRUCTED), "SEQUENCE");
   return BER_Decoder(object.value, object.value.size());
   }

}

}
